#include "stdafx.h"
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <string.h>
#include "WinModFileTypes.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// Zu finden in Mod_Loader.cpp
extern int _read16BITof68000(int h, unsigned short *word);
extern int _read32BITof68000(int h, unsigned long *dword);
extern int _readNormName(int h, char *buffer, unsigned long len);
extern char ModExtension[4];

void Set_StmType(int h, struct TheModFile &ModClass)
{
	char Kennung[9];

	// Kurz die Kennung lesen:
	*Kennung=0;
	lseek(h,0x14L,SEEK_SET);
	_read(h,&Kennung,8L); Kennung[8]=0;
	lseek(h,0L,SEEK_SET);

	// Modtyp erkennen:
	ModClass.nSamples=31;
	ModClass.nChannels=4;
	strcpy(ModClass.TrackerName, "ScreamTracker");
}

void translate(unsigned char *oldEffekt, unsigned char *oldEffektByte)
{
	unsigned char efftrans[] = {0x00,0x0F,0x0B,0x0D,0x0A,0x02,0x01,0x03,0x04,0x07,0x00,0x08};

	if (*oldEffekt>=0 && *oldEffekt<=11) *oldEffekt=efftrans[*oldEffekt];
	if (*oldEffekt==0x0F)
		*oldEffektByte=(unsigned char)(((*oldEffektByte)&0xF0)>>4) | (unsigned char)(((*oldEffektByte)&0x0F)<<4);
}

void InsertNote_Stm(unsigned int PattNum, unsigned int TimeIndex, unsigned int ChannelIndex, unsigned long note, struct TheModFile &ModClass)
{
	struct PatternData *p,*lauf;
	struct PatternData *root;

	root=ModClass.Pattern[PattNum];
	p=new struct PatternData;

	p->TimeIndex=TimeIndex;
	p->ChannelIndex=ChannelIndex;
	
	p->SampleIndex=(note&0xF80000L)>>19;
	int Oktave = (note&0xF0000000L)>>28;
	int Ton =    (note&0x0F000000L)>>24;
	int index = (Oktave+4)*12+Ton;
	if (index<nNotes) p->Ton=notes[index].NoteVal; else p->Ton=0;

	int hlp=GetNoteIndex(p->Ton);
	p->noteIndex = hlp%12;
	p->noteOctave = hlp/12;

	p->Effekt=(unsigned char)((note&0xF00)>>8);
	p->Effektop=(unsigned char)(note&0xFF);
	translate(&p->Effekt, &p->Effektop);
	
	p->Volume=((note&0x70000)>>16) | ((note&0xF000)>>9);
	if (p->Volume>64) p->Volume=-1;
	
	p->Next=p->Prev=NULL;

	if (!root) ModClass.Pattern[PattNum]=p;
	else
	{
		lauf=root;
		while (lauf)
		{
			if (!lauf->Next) break;
			if (lauf->Next->TimeIndex>TimeIndex) break;
			lauf=lauf->Next;
		}
		p->Next=lauf->Next;
		p->Prev=lauf;
		if (lauf->Next) lauf->Next->Prev=p;
		lauf->Next=p;
	}
}
int Read_Stm(int h, struct TheModFile &ModClass, int *Error, long *ErrorValue)
{
	char ReadNames[40];
	unsigned long ReadDWord;
	unsigned short ReadWord;
	unsigned char ReadByte;
	unsigned int i, Channel, row, PattNum;

	// Angenommen!
	*Error=LOADERROR_NOERROR;
	
	ModClass.usesLinearFrequency = false;	// Standard bei STM
	// Modtyp feststellen
	Set_StmType(h, ModClass);

	// Liest den Songnamen inclusive abschiesendem Null-Byte
	_readNormName(h, ReadNames, 20L); ReadNames[20]=0;
	strcpy(ModClass.SongName,ReadNames);

	// Die Kennung
	_read(h, ReadNames, 8L); ReadNames[8]=0;
	strcpy(ModClass.ModID, ReadNames);

	// 0x1A in der Regel: Nur zum berlesen!
	_read(h, &ReadByte, 1L);

	// FileType: 1 = Song, 2 = Mod
	_read(h, &ReadByte, 1L);
	ModClass.FileType = ReadByte;
	if (ReadByte!=2)
	{
		*Error=LOADERROR_UNSUPPORTED;
		return(0);
	}

	// Version:
	_read(h, &ReadWord, 2L);
	sprintf(ModClass.TrackerName, "%s V%d.%d", ModClass.TrackerName, (int)(ReadWord&0x0F), (int)(ReadWord>>8));

	// PlaybackTempo
	_read(h, &ReadByte, 1L);
	ModClass.Tempo = 6;
	ModClass.BPMSpeed = 125;

	// Songlnge
	_read(h, &ReadByte, 1L);
	if (ReadByte>64) 
	{
		ReadByte=64;
		*Error=LOADERROR_MODCORRUPT;
	}
	ModClass.nPattern=(unsigned int)ReadByte;

	// Playback Volume
	_read(h, &ReadByte, 1L);
	ModClass.BaseVolume = ReadByte;

	// reserved
	_read(h, ReadNames, 13L);

	// Samples lesen
	ModClass.Samples=new struct SampleData [ModClass.nSamples];
	for (i=0; i<ModClass.nSamples; i++)
	{
		// Samplename
		_readNormName(h, ReadNames, 12L); ReadNames[12]=0;
		strcpy(ModClass.Samples[i].Name, ReadNames);

		// Reserved
		_read(h, &ReadByte, 1L);

		// InstrumentDisk, falls Song:
		_read(h, &ReadByte, 1L);
		if (ModClass.FileType == 1) sprintf(ModClass.Samples[i].Name, "%s, #%d", ModClass.Samples[i].Name, (int)(ReadByte));

		// Reserved (Sample Beginning Offset?!)
		_read(h, &ReadWord, 2L);

		// Anzahl Sampldaten
		_read(h, &ReadWord, 2L);
		ModClass.Samples[i].Length=((unsigned long)ReadWord);

		// RepeatStart
		_read(h, &ReadWord, 2L);
		ModClass.Samples[i].RepeatStart=(unsigned long)ReadWord;

		// RepeatStop
		_read(h, &ReadWord, 2L);
		ModClass.Samples[i].RepeatStop=(unsigned long)ReadWord; //+ ModClass.Samples[i].RepeatStart;

		// No Loop with these Values!
		if ((ModClass.Samples[i].RepeatStart==ModClass.Samples[i].RepeatStop) || (!ModClass.Samples[i].RepeatStart && ModClass.Samples[i].RepeatStop==2) || ModClass.Samples[i].RepeatStop==0xFFFF)
		{
			ModClass.Samples[i].RepeatStart=0;
			ModClass.Samples[i].RepeatStop=0;
		}
		else
		// Begrenzen auf Samplelnge, falls nicht korrekt. (sicher ist sicher!)
		if (ModClass.Samples[i].RepeatStop>ModClass.Samples[i].Length) 
			ModClass.Samples[i].RepeatStop=ModClass.Samples[i].Length;

		// Lautstrke
		_read(h, &ReadByte, 1L);
		ModClass.Samples[i].Volume=ReadByte;

		// Reserved
		_read(h, &ReadByte, 1L);

		// BaseFrequ:
		_read(h, &ReadWord, 2L);

		// Restlichen Werte:
		ModClass.Samples[i].FineTune=0;
		ModClass.Samples[i].BaseFrequenzy = ReadWord;
		ModClass.Samples[i].BaseTone = getTonToFrequenzy(ModClass.Samples[i].BaseFrequenzy);

		// Reserved
		_read(h, &ReadDWord, 4L);

		// Length in Paragraphs. Ignoring:
		_read (h, &ReadWord, 2L);
	}

	// Arrangement
	ModClass.SongArrangement=new unsigned long[128];
	ModClass.SongLength = 99;
	for (i=0; i<128; i++)
	{
		_read(h, &ReadByte, 1L);
		ModClass.SongArrangement[i]=ReadByte;
		if (ModClass.SongLength==99 && ReadByte==99) ModClass.SongLength = i;
	}
	
	while (ModClass.SongArrangement[ModClass.SongLength-1]>=ModClass.nPattern) ModClass.SongLength--;

	// Speicher fr alle Pattern reservieren und dann auf NULL initialisieren
	ModClass.Pattern=(struct PatternData **)new (struct PatternData (*[ModClass.nPattern]));
	for (i=0; i<ModClass.nPattern; i++) ModClass.Pattern[i]=NULL;

	for (PattNum=0; PattNum<ModClass.nPattern; PattNum++)
		for (row=0; row<64; row++)
			for (Channel=0; Channel<ModClass.nChannels; Channel++)
			{
				_read32BITof68000(h, &ReadDWord);
				if (ReadDWord) InsertNote_Stm(PattNum, row, Channel, ReadDWord, ModClass);
			}

	for (i=0; i<ModClass.nSamples; i++)
	{
		unsigned long s, SamLen;
		signed char *TempBuffer;

		SamLen=ModClass.Samples[i].Length;
		if (SamLen)
		{
			ModClass.Samples[i].Sample=new signed short[SamLen];
			TempBuffer=new signed char[SamLen];
			_read(h, TempBuffer, SamLen);

			//ModClass.Samples[i].MaxDiff=0;
			for (s=0; s<SamLen; s++)
			{
				ModClass.Samples[i].Sample[s]=PromModTo16Bit(TempBuffer[s]);
			}

			delete [] TempBuffer;
		} 
		else 
			ModClass.Samples[i].Sample=NULL;
	}

	ModClass.DefaultPanning=0;
	ModClass.panningValuesL = new signed long[ModClass.nChannels];
	ModClass.panningValuesR = new signed long[ModClass.nChannels];
	
	for (i=0; i<ModClass.nChannels; i++)
	{
		// Fr Stereo, was aber STMs eigentlich nicht sind, daher:
		//ModClass.panningValuesR[i]=LMULTI; ModClass.panningValuesL[i]=LMULTI;
		if (i%3)
		{
			ModClass.panningValuesR[i]=LMULTI; 
			ModClass.panningValuesL[i]=0;
		}
		else
		{
			ModClass.panningValuesL[i]=LMULTI; 
			ModClass.panningValuesR[i]=0;
		}
	}

	return(1); // Success
}
