#include <mmsystem.h>
#include <mmreg.h>
#include <dsound.h>
#include "WinMod.h"
#include "WinModDlg.h"
#include "WinModInsDlg.h"
#include "WinModPatternDlg.h"
// WinModFileTypes:
// Stellt alle ntigen Strukturen und Variablen der Classe WinMod FileType zur Verfgung

// Formellehre:
// (BaseTone*BaseFreq)/(ton*PlayerFrequ) == Faktor fr Sampledaten

// Durch Ergnzung mit LMULTI (Ergebnis genauer) bedeutet dies:
// Solange errechnetem Wert aufaddieren, bis >LMULTI (>1), 
// dann LMULTI abziehen (-1) und ein Sample weiter.
// [Zwischenzeitlich ist lineare (cubische?!?) Interpolation gefragt!]

// (25*PlayerFrequ)/(BPMOp*10) == Anzahl Sampledaten zwischen den Aufrufen zu GoNextNote
// wobei (25*PlayerFrequ)/10 bereits ausgerechnet sein knnte. Runden ist erlaubt.
// "Set Speed" bedeutet, da zwischen zwei Patternrowabrufen die Angegeben
// Anzahl Ticks kommen.
// Also bei 125 BPM und Speed 6==> bei 44100Hz: 882 Sample (pro Kanal!),
// dann GoNextNote, das wiederrum 5 Ticks warten mu. Nchste Note auf der 6.

#define QualityShift (12) //4
#define QualityShiftBack (4)  //4

#define ClippingValuePos ((signed long) (8388607L)) // ((signed long) (32767L))
#define ClippingValueNeg ((signed long)(-8388608L)) // ((signed long)(-32768L))

#define DeClickBorder ((signed long)(4096000))		// ((signed long)(16000))

#define nComboFrequenzies		7
#define nComboBitPerSamples		3
#define nComboChannels			2
#define nComboBufferms			11
#define DefaultBufferms			1
#define DefaultFrequenzy		3
#define DefaultChannel			1
#define DefaultSamples			1

#define NUM_PLAY_NOTIFICATIONS  8

#define swapword(word)			((((word) & 0xFF)<<8) | ((word)>>8))
#define swapdword(dword)		(swapword(dword&0xFFFF)<<16) | swapword(dword>>16)

#define PromModTo16Bit(aByte)	(((signed int)aByte)<<8)
#define PromoteTo16Bit(aByte)	(((signed int)aByte-128)<<8)
#define PromoteTo8Bit(aWord)	((unsigned char)((aWord>>8)+128))

#define LOADERROR_NOERROR		0
#define LOADERROR_MODSIZE		1
#define LOADERROR_UNSUPPORTED	2
#define LOADERROR_FILENOTFOUND	3
#define LOADERROR_MODCORRUPT	4
#define LOADERROR_MODANZCHAN	5

#define BASETone				((unsigned long)(428))		// C-2
#define BASEFreq				((unsigned long)(8363))		// Frequenz

#define LMUTLISHIFT				10
#define LMULTI					((unsigned long)(1<<LMUTLISHIFT))		// Multiplikator
#define FMULTI					((double)(LMULTI))

#define nNotes					132

#define VOLUMEDL(c,v)			(signed long)((((__int64)((c->ActVolumeL*ThisMod->BaseVolume)/MaxVolValue))*((__int64)(v)))/((__int64)(MaxVolume)))
#define VOLUMEDR(c,v)			(signed long)((((__int64)((c->ActVolumeR*ThisMod->BaseVolume)/MaxVolValue))*((__int64)(v)))/((__int64)(MaxVolume)))
#define MaxVolValue				((signed long)(64L))
#define StartingVolume			((signed long)(ThisMod->BaseVolume*((ThisMod->nChannels<=4)?(ThisMod->nChannels):(ThisMod->nChannels>>1))))

#define NoteCutValue			(unsigned long)(-1)

#define WINMOD_REG_KEY			TEXT("Software\\QuippySoft\\WinMod")

struct SampleData
{
	char Name[40];
	unsigned long Length;		// Bereits mit zwei multipliziert!
	signed char FineTune;		// Finetuning -8..+8
	signed long Volume;			// Basisvolume
	//signed long MaxDiff;		// Nimm die maximale Differenz zweier Sample in diesem Instrument auf
	unsigned long RepeatStart;	// Wert des Startpunktes (bereits mal 2)
	unsigned long RepeatStop;	// Wert des Endpunktes   (Bereits mal 2)
	unsigned long BaseTone;		// Basiston
	unsigned long BaseFrequenzy;// BaseFrequenzy
	//S3M:
	unsigned char Type;			// Sollte 1 sein fr Sample
	char DosFileName[13];		// DOS File-Name
	unsigned char Flags;		// Loop an/aus
	unsigned short MemSeg;		// Position des Samples in der Datei...
	signed short *Sample;		// Das Sample selbst. Die 8-Bit werden automatisch in 16-Bit konvertiert.
								// 8Bit: 0..127,128-255; 16Bit: -32768..0..+32767
};

struct PatternData
{
	unsigned long TimeIndex;	// Hier steht die aktuelle Row. Es ergibt sich eine doppelt Verkettete Liste
	unsigned long ChannelIndex;	// Stimme
	unsigned long Ton;			// Die Tonhhe
	unsigned char noteIndex;	// Der Tonindex
	unsigned char noteOctave;	// Die Oktave
	unsigned long SampleIndex;	// Das zu spielende Sample
	unsigned char Effekt;		// Das Effektbyte
	signed long Volume;			// aktuelles Volume (0-64), -1== NoEffekt
	unsigned char Effektop;		// Beide zusammen
	struct PatternData *Next;	// Verkettete Liste
	struct PatternData *Prev;	// und zwar doppelt!
};

struct TheModFile
{
	char SongName[40];
	char TrackerName[40];

	struct SampleData *Samples;

	unsigned long	nSamples;			// Wird durch den Modtyp festgelegt
	unsigned long	nChannels;			// Wird durch den Modtyp festgelegt
	unsigned long	SongLength;			// Anzahl der zu spielenden Pattern
	unsigned long	nPattern;			// Anzahl der im Mod gespeicherten Pattern
	unsigned short	CIAA;				// CIAA-Speed, meistens 127
	unsigned long	*SongArrangement;	// Das Songarrangement, die maximale Anzahl hngt vom modtyp ab.
	unsigned int	BPMSpeed;			// BPM
	unsigned int	Tempo;				// Tempo
	unsigned long	GlobalTuning;		// Vorausgerechnetes Tuning
	// Fr STM / S3M
	unsigned char	FileType;			// 1=Song, 2=Module. Nur 2 ist supported!
	unsigned char	BaseVolume;			// Basisvolume.
	// Fr S3M
	unsigned char	SampleType;			// 1: signed, 2:unsigned
	unsigned char	DefaultPanning;		// Wenn 252 (0xFC), dann gibt es PanningData
	signed long		*panningValuesR;	// Wird vom Loader initialisiert. Werte gehen von -128 - +127 *LMULTI
	signed long		*panningValuesL;	// Wird vom Loader initialisiert. Werte gehen von -128 - +127 *LMULTI

	char ModID[10];						// Die ModID

	struct PatternData **Pattern;		// Die Rootzeiger auf die einzelnen Pattern.
										// Pattern[0] ist z.B. Rootzeiger auf die Liste des Pattern 0
	bool usesLinearFrequency;			// Setzt, ob lineare Frequenzen benutzt werden sollen.
};

//Externe Deklarationen
extern struct TheModFile *ThisMod;

extern struct noteTranslation 
{
   char NoteStr[4];
   unsigned long NoteVal;
} notes[];

extern unsigned int Frequenzies[nComboFrequenzies];
extern unsigned int Channels[nComboChannels];
extern unsigned int BitPerSamples[nComboBitPerSamples];
extern unsigned int Bufferms[nComboBufferms];
extern char *version;

// Allgemeine Prototypen:
char			*GetNote(unsigned long nVal);
int				GetNoteIndex(unsigned long nVal);
unsigned long	RoundTon(unsigned long nVal);
unsigned long	getTonToFrequenzy(unsigned long Freq);
void			FreeModStruct(struct TheModFile **TheMod);
void			GenModStruct(struct TheModFile **TheMod);
int				Read_ModFiles(char *FileName, struct TheModFile &ModClass, int *Error, long *ErrorValue);

// Directsound Prototypen
extern LPDIRECTSOUND		g_pDS;
extern LPDIRECTSOUNDBUFFER	g_pDSBuffer;
extern LPDIRECTSOUNDNOTIFY	lpdsNotify;
extern DSBPOSITIONNOTIFY	rgdsbpn[NUM_PLAY_NOTIFICATIONS];
extern HANDLE				rghEvent[NUM_PLAY_NOTIFICATIONS];
extern DSBUFFERDESC         dsbd;
extern HWAVEOUT				hWaveOut; // Handle to the Waveout, if used
extern bool					useWaveOut;
extern WAVEHDR				*blocks;
HRESULT	InitDirectSound( HWND hDlg, int nSamplesPerSec, int nBitsPerSample, int nChannels, DWORD WaveCallBack);
HRESULT FreeDirectSound();
HRESULT CreateStreamingBuffer(int nSamplesPerSec, int nBitsPerSample, int nChannels, bool Notify);
HRESULT StopBuffer();
HRESULT StartBuffer();
HRESULT PauseBuffer(bool Pause);
int RetrieveActualNotifySegment();

//Player Prototypen
extern unsigned long		FineTuneTable[16];
extern signed long			VibratoTable[32];
extern unsigned int			PlayerFrequ;
extern unsigned int			PlayerBitPerSample;
extern unsigned int			PlayerChannels;
extern unsigned int			PlayerBufferLenMillisec;
extern bool					PlayerISP;
extern bool					PlayerClipping;
extern bool					PlayerLoopsAllowed;
extern bool					PlayerDone;
extern bool					PlayerWideStereo;
extern bool					PlayerDigiReady;
extern bool					PlayerClick;
extern bool					PlayerFastForward;
extern bool					PlayerRewind;
extern unsigned long		PlayerActRow, PlayerActPattern, PlayerActPatternIndex;
extern signed long			MaxVolume;
extern unsigned int			PlayerActTicks;
extern unsigned long		BPMSpeed, BufferDataLen;
extern unsigned short		Tempo;
extern signed short			*StereoTabL, *StereoTabR;
extern unsigned long		PlayerMilliSec;
extern bool					PlayerUseLinearFrequency;
void						PlayThePiece(HWND hWnd);
void						StopPlay();
void						DoFastForward(bool anaus);
void						DoRewind(bool anaus);

struct voicememory
{
	bool				PlayerMute,VolumeDoItFine, PortaDoItFine, Glissando, 
						VolSlideOn, InstrHit;
	unsigned long		SampleCount, SamplePos, PatternTon, ActTon, PlayerTuning, NoteDelayTon;
	signed long			ActVolume,ActVolumeL,ActVolumeR, VolSlide, PanningValueR, PanningValueL, Olds1;
	unsigned char		Retrig, RetRetrig, NoteDelay, NoteCut, TheEffektByte, TheEffektOp, PatternNoteIndex, PatternNoteOctave;
	struct SampleData	*PlayerActSampleData, *NoteDelayInstrument;
	
	struct TuneNote
	{
		bool PortaOn;
		signed long Step, OldPortaStep, OldPortaNoteStep, OldFinePortaStep;
		signed long ActNote,StartNote,EndNote;
	} Porta;

	struct VibratoNote
	{
		bool VibratoOn, Negativ;
		unsigned long ActTablePos;
		signed long Step;
		signed long Amplitude;
		unsigned char Type;
	} Vibrato;

	struct TremoloNote
	{
		bool TremoloOn, Negativ;
		unsigned long ActTablePos;
		signed long Step;
		signed long Amplitude;
		unsigned char Type;
	} Tremolo;

	struct Arpeggio
	{
		bool ArpOn;
		int Index;
		signed long Note[3];
	} Arp;
};

extern struct voicememory *voiceeffekts;

// Effektroutinen anlegen.
extern void DoEffekt0(struct voicememory *, bool StartEffekt);
extern void DoEffekt1(struct voicememory *, bool StartEffekt);
extern void DoEffekt2(struct voicememory *, bool StartEffekt);
extern void DoEffekt3(struct voicememory *, bool StartEffekt);
extern void DoEffekt4(struct voicememory *, bool StartEffekt);
extern void DoEffekt5(struct voicememory *, bool StartEffekt);
extern void DoEffekt6(struct voicememory *, bool StartEffekt);
extern void DoEffekt7(struct voicememory *, bool StartEffekt);
extern void DoEffekt8(struct voicememory *, bool StartEffekt);
extern void DoEffekt9(struct voicememory *, bool StartEffekt);
extern void DoEffektA(struct voicememory *, bool StartEffekt);
extern void DoEffektB(struct voicememory *, bool StartEffekt);
extern void DoEffektC(struct voicememory *, bool StartEffekt);
extern void DoEffektD(struct voicememory *, bool StartEffekt);
extern void DoEffektE(struct voicememory *, bool StartEffekt);
extern void DoEffektF(struct voicememory *, bool StartEffekt);
extern void DoEffektQ(struct voicememory *, bool StartEffekt);
extern void DoEffekt10(struct voicememory *, bool StartEffekt);
extern void DoEffektDummy(struct voicememory *, bool StartEffekt);
extern void (*DoEffekts[26])(struct voicememory *, bool StartEffekt);
